%   * dpp_train_2class.m
%   * Copyright (C) Manuel Fernandez Delgado 2011 <manuel.fernandez.delgado@usc.es>
%   *
%   *     This program is free software: you can redistribute it and/or modify
%   *     it under the terms of the GNU General Public License as published by
%   *     the Free Software Foundation, either version 3 of the License, or
%   *     (at your option) any later version.
%   *
%   *     This program is distributed in the hope that it will be useful,
%   *     but WITHOUT ANY WARRANTY; without even the implied warranty of
%   *     MERCHANPOILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%   *     GNU General Public License for more details.
%   *
%   *     You should have received a copy of the GNU General Public License
%   *     along with this program.  If not, see <http://www.gnu.org/licenses/>.
%   *

function [w ps] = dpp_train_2class(x, d, n, n_init)
% DPP_TRAIN trains a two-class DPP network with n parallel perceptrons
%  Syntax:
%      [w ps] = dpp_train_2class(x, d, n, n_init)
%  Description
%  x - (N x ni) matrix with N training patterns of dimension ni (no. inputs)
%  d - (Nx1) matrix with the desired output class (0 or 1)
%  n - number of parallel perceptrons
%  n_init - number of initializations: the w with highest accuracy on the training set is selected
%  Returned values
%  w - (n x ni) weight vectors for the n perceptrons
%  ps - process settings calculated from the training patterns
% Notation refered to page 1841 of paper "Direct Parallel Perceptrons (DPPs): Fast Analytical 
% Calculation of the Parallel Perceptrons Weights with Margin Control for
% Classiﬁcation Tasks", M. Fernández-Delgado, J. Ribeiro, E. Cernadas, and
% S. Barro, IEEE Trans. on Neural Networks, Vol. 22, No. 11, November 2011
[x ps] = mapstd(x'); x = x'; % pre-processing
N = length(d);  % N = no. training patterns
if any(d ~= 0 & d ~= 1)
    error('argument d must be 0 or 1')
end
d(d == 0) = -1;
x = [x ones(N,1)]; 
ni = size(x,2);
m = fix((n - 1)/2);
xi = zeros(1,m);
dik = zeros(n, N);
acc_best = 0;
for init = 1:n_init   % initialization loop:
    for i = 1:n
        dik(i, :) = d;
    end
    % generate dik
    for i=1:N
        for j=1:m
            while 1     % avoids repeated numbers in vector xi
                xi(j) = randi(n,1);
                if all(xi(j) ~= xi(1:j-1))
                    break
                end
            end
        end
        dik(xi,i) = - dik(xi,i);
    end
    w = zeros(n, ni);
    for i = 1:n
        for j=1:N
            w(i,:) = w(i,:) + dik(i,j)*x(j,:);
        end
        w(i,:) = w(i,:)/norm(w(i,:));
    end
    acc = 0;
    for i = 1:N
        y = 0;
        for j = 1:n
            y = y + sign(dot(w(j,:),x(i,:)));
        end
        y = sign(y);
        if y == d(i)
            acc = acc + 1;
        end
    end
    acc = 100*acc/N;
    if acc > acc_best
        w_best = w;
        acc_best = acc;
    end
    fprintf('init %i acc= %.1f%%\n', init, acc);
end
w = w_best;
fprintf('acc_best= %.1f%%\n', acc_best);
end % ends function 
