%   * dpp_train_multiclass.m
%   * Copyright (C) Manuel Fernandez Delgado 2011 <manuel.fernandez.delgado@usc.es>
%   *
%   *     This program is free software: you can redistribute it and/or modify
%   *     it under the terms of the GNU General Public License as published by
%   *     the Free Software Foundation, either version 3 of the License, or
%   *     (at your option) any later version.
%   *
%   *     This program is distributed in the hope that it will be useful,
%   *     but WITHOUT ANY WARRANTY; without even the implied warranty of
%   *     MERCHANPOILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%   *     GNU General Public License for more details.
%   *
%   *     You should have received a copy of the GNU General Public License
%   *     along with this program.  If not, see <http://www.gnu.org/licenses/>.
%   *
function [w ps] = dpp_train_multiclass(x, d, n, n_init)
% DPP_TRAIN_MULTICLASS trains a multiclass DPP network with n parallel perceptrons
%  Syntax:
%      [w ps] = dpp_train_multiclass(x, d, n, n_init)
%  Description
%  x - (Nxni) matrix with N training patterns of dimension ni (no. inputs)
%  d - Nx1 matrix with desired output class (1..number of classes)
%  n - number of parallel perceptrons
%  n_init - number of random initializations: the w with highest accuracy on the training set is selected
%  Returned values
%  w - (nc*n x ni) matrix, weight vectors for the n perceptrons of the nc DPPs
%  ps - process settings calculated from the training patterns
% Notation refered to page 1841 of paper "Direct Parallel Perceptrons (DPPs): Fast Analytical 
% Calculation of the Parallel Perceptrons Weights with Margin Control for
% Classiﬁcation Tasks", M. Fernández-Delgado, J. Ribeiro, E. Cernadas, and
% S. Barro, IEEE Trans. on Neural Networks, Vol. 22, No. 11, November 2011
if min(d) ~= 1
    error('min(d) must be 1')
end
nc = max(d);    % nc = no. of classes
N = length(d);  % N = no. training patterns
for i=1:N
    if all(d(i) ~= 1:nc)
        disp(d)
        error('d must be in 1..nc')
    end
end
s = zeros(1, nc);
dlk = zeros(1, N); 
dik = zeros(n, N);
[x ps] = mapstd(x'); x = x'; % pre-processing
x = [x ones(N,1)];  % concatenates the constant offset (=1) input
ni = size(x,2);
w = zeros(nc*n, ni);
m = fix((n - 1)/2);
xi = zeros(1,m);
acc_best = 0;
for init = 1:n_init   % initialization loop
    for c = 1:nc    % loop which runs the nc DPPs (one for each class)
        for j = 1:N
            if d(j) == c
                dlk(j) = 1;
            else
                dlk(j) = -1;
            end
        end
        for j = 1:n
            dik(j,:) = dlk;
        end
        for i=1:N
            for j=1:m
                while 1     % avoids repeated numbers in vector xi
                    xi(j) = randi(n,1);
                    if all(xi(j) ~= xi(1:j-1))
                        break
                    end
                end
            end
            dik(xi,i) = - dik(xi,i);
        end
        for i = 1:n
            index = (c - 1)*nc + i;
            w(index, :) = 0;
            for j=1:N
                w(index,:) = w(index,:) + dik(i,j)*x(j,:);
            end
            w(index,:) = w(index,:)/norm(w(index,:));
        end
    end
    acc = 0;
    for i = 1:N
        vmax = -inf;
        for c = 1:nc
            s(c) = 0;
            for j = 1:n
                index = (c - 1)*nc + j;
                s(c) = s(c) + dot(w(index,:),x(i,:));
            end
            if s(c) > vmax
                vmax = s(c); y = c;
            end
        end
        if y == d(i)
            acc = acc + 1;
        end
    end
    acc = 100*acc/N;
    if acc > acc_best
        w_best = w;
        acc_best = acc;
    end
    fprintf('init %i acc= %.1f%%\n', init, acc);
end
w = w_best;
fprintf('acc_best= %.1f%%\n', acc_best);
end % ends function 
